package lbms.plugins.mldht.kad;

import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;

import lbms.plugins.mldht.kad.messages.AnnounceRequest;
import lbms.plugins.mldht.kad.messages.AnnounceResponse;
import lbms.plugins.mldht.kad.messages.ErrorMessage;
import lbms.plugins.mldht.kad.messages.FindNodeRequest;
import lbms.plugins.mldht.kad.messages.FindNodeResponse;
import lbms.plugins.mldht.kad.messages.GetPeersRequest;
import lbms.plugins.mldht.kad.messages.GetPeersResponse;
import lbms.plugins.mldht.kad.messages.MessageBase;
import lbms.plugins.mldht.kad.messages.PingRequest;
import lbms.plugins.mldht.kad.messages.PingResponse;

/**
 * @author Damokles
 *
 */
public class DHT implements DHTBase {

	private static DHTLogger				logger;
	private static LogLevel					logLevel	= LogLevel.Info;

	private static ScheduledExecutorService	scheduler;
	private static ThreadGroup				executorGroup;

	static {
		executorGroup = new ThreadGroup("mlDHT");
		scheduler = Executors.newScheduledThreadPool(2, new ThreadFactory() {
			public Thread newThread (Runnable r) {
				Thread t = new Thread(executorGroup, r, "mlDHT Executor");

				t.setDaemon(true);
				return t;
			}
		});

		logger = new DHTLogger() {
			public void log (String message) {
				System.out.println(message);
			};

			/* (non-Javadoc)
			 * @see lbms.plugins.mldht.kad.DHTLogger#log(java.lang.Exception)
			 */
			public void log (Exception e) {
				e.printStackTrace();
			}
		};
	}

	private boolean							running;

	private boolean							bootstrapping;
	private long							lastBootstrap;

	private int								port;
	private Node							node;
	private RPCServerBase					srv;
	private Database						db;
	private TaskManager						tman;
	private ExpireTimer						expire_timer;
	private File							table_file;

	private List<DHTStatsListener>			statsListeners;
	private List<DHTStatusListener>			statusListeners;
	private DHTStats						stats;
	private DHTStatus						status;

	public DHT () {
		expire_timer = new ExpireTimer();
		stats = new DHTStats();
		status = DHTStatus.Stopped;
		statsListeners = new ArrayList<DHTStatsListener>(2);
		statusListeners = new ArrayList<DHTStatusListener>(2);

	}

	public void ping (PingRequest r) {
		if (!running) {
			return;
		}

		// ignore requests we get from ourself
		if (r.getID().equals(node.getOurID())) {
			return;
		}

		PingResponse rsp = new PingResponse(r.getMTID(), node.getOurID());
		rsp.setOrigin(r.getOrigin());
		srv.sendMessage(rsp);
		node.recieved(this, r);
	}

	public void findNode (FindNodeRequest r) {
		if (!running) {
			return;
		}

		// ignore requests we get from ourself
		if (r.getID().equals(node.getOurID())) {
			return;
		}

		node.recieved(this, r);
		// find the K closest nodes and pack them
		KClosestNodesSearch kns = new KClosestNodesSearch(r.getTarget(),
				DHTConstants.MAX_ENTRIES_PER_BUCKET);

		node.findKClosestNodes(kns);

		FindNodeResponse fnr = new FindNodeResponse(r.getMTID(),
				node.getOurID(), kns.pack());
		fnr.setOrigin(r.getOrigin());
		srv.sendMessage(fnr);
	}

	public void response (MessageBase r) {
		if (!running) {
			return;
		}

		node.recieved(this, r);
	}

	public void getPeers (GetPeersRequest r) {
		if (!running) {
			return;
		}

		// ignore requests we get from ourself
		if (r.getID().equals(node.getOurID())) {
			return;
		}

		node.recieved(this, r);
		List<DBItem> dbl = new ArrayList<DBItem>();
		db.sample(r.getInfoHash(), dbl, 50);

		// generate a token
		Key token = db.genToken(r.getOrigin().getAddress().getAddress(),
				r.getOrigin().getPort());

		if (dbl.size() == 0) {
			// if data is null do the same as when we have a findNode request
			// find the K closest nodes and pack them
			KClosestNodesSearch kns = new KClosestNodesSearch(r.getInfoHash(),
					DHTConstants.MAX_ENTRIES_PER_BUCKET);
			node.findKClosestNodes(kns);

			GetPeersResponse fnr = new GetPeersResponse(r.getMTID(),
					node.getOurID(), kns.pack(), token);
			fnr.setOrigin(r.getOrigin());
			srv.sendMessage(fnr);
		} else {
			// send a get peers response
			GetPeersResponse fvr = new GetPeersResponse(r.getMTID(),
					node.getOurID(), dbl, token);
			fvr.setOrigin(r.getOrigin());
			srv.sendMessage(fvr);
		}
	}

	public void announce (AnnounceRequest r) {
		if (!running) {
			return;
		}

		// ignore requests we get from ourself
		if (r.getID().equals(node.getOurID())) {
			return;
		}

		node.recieved(this, r);
		// first check if the token is OK
		Key token = r.getToken();
		if (!db.checkToken(token, r.getOrigin().getAddress().getAddress(),
				r.getOrigin().getPort())) {
			logDebug("DHT Received Announce Request with invalid token.");
			return;
		}

		logDebug("DHT Received Announce Request, adding peer to db: "
				+ r.getOrigin().getAddress());

		// everything OK, so store the value
		byte[] tdata = new byte[6];
		ByteBuffer bb = ByteBuffer.wrap(tdata);
		bb.put(r.getOrigin().getAddress().getAddress(), 0, 4);
		bb.putShort((short) r.getOrigin().getPort());
		db.store(r.getInfoHash(), new DBItem(tdata));
		// send a proper response to indicate everything is OK
		AnnounceResponse rsp = new AnnounceResponse(r.getMTID(),
				node.getOurID());
		rsp.setOrigin(r.getOrigin());
		srv.sendMessage(rsp);
	}

	public void error (ErrorMessage r) {

	}

	public void timeout (MessageBase r) {
		if (running) {
			node.onTimeout(r);
		}
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#addDHTNode(java.lang.String, int)
	 */
	public void addDHTNode (String host, int hport) {
		if (!running) {
			return;
		}
		InetSocketAddress addr = new InetSocketAddress(host, hport);

		if (!addr.isUnresolved()) {
			srv.ping(node.getOurID(), addr);
		}

	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#announce(byte[], int)
	 */
	public AnnounceTask announce (byte[] info_hash, int port) {
		if (!running) {
			return null;
		}
		Key id = new Key(info_hash);

		AnnounceTask at = new AnnounceTask(db, srv, node, id, port);
		if (canStartTask()) {
			at.start();
		}
		tman.addTask(at);
		if (!db.contains(id)) {
			db.insert(id);
		}
		return at;

	}

	public PingRefreshTask refreshBuckets (KBucket[] buckets,
			boolean cleanOnTimeout) {
		PingRefreshTask prt = new PingRefreshTask(srv, node, buckets,
				cleanOnTimeout);
		if (canStartTask()) {
			prt.start();
		}
		tman.addTask(prt);
		return prt;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#getClosestGoodNodes(int)
	 */
	public Map<String, Integer> getClosestGoodNodes (int maxNodes) {
		Map<String, Integer> map = new HashMap<String, Integer>();

		if (node == null) {
			return map;
		}

		int max = 0;
		KClosestNodesSearch kns = new KClosestNodesSearch(node.getOurID(),
				maxNodes * 2);
		node.findKClosestNodes(kns);

		for (KBucketEntry e : kns.getEntries()) {

			if (!e.isGood()) {
				continue;
			}

			InetSocketAddress a = e.getAddress();

			map.put(a.getHostName(), a.getPort());
			if (++max >= maxNodes) {
				break;
			}
		}
		return map;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#getPort()
	 */
	public int getPort () {
		return port;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#getStats()
	 */
	public DHTStats getStats () {
		return stats;
	}

	/**
	 * @return the status
	 */
	public DHTStatus getStatus () {
		return status;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#isRunning()
	 */
	public boolean isRunning () {
		return running;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#portRecieved(java.lang.String, int)
	 */
	public void portRecieved (String ip, int port) {
		if (!running) {
			return;
		}

		PingRequest r = new PingRequest(node.getOurID());
		r.setOrigin(new InetSocketAddress(ip, port));
		srv.doCall(r);

	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#start(java.lang.String, int)
	 */
	public void start (File table, int port) throws SocketException {
		if (running) {
			return;
		}

		if (port == 0) {
			port = 49001;
		}
		setStatus(DHTStatus.Initializing);
		stats.resetStartedTimestamp();

		table_file = table;
		this.port = port;
		logInfo("Starting DHT on port " + port);
		srv = new RPCServer(this, port);
		stats.setRpcStats(srv.getStats());
		node = new Node(srv);
		db = new Database();
		stats.setDbStats(db.getStats());
		tman = new TaskManager();
		expire_timer.update();
		running = true;
		srv.start();
		bootstrapping = true;
		node.loadTable(table,this,new Runnable() {
			public void run() {
				bootstrapping = false;
				bootstrap();
			}
		});
		scheduler.scheduleAtFixedRate(new Runnable() {
			public void run () {
				try
				{
					update();
				} catch (RuntimeException e)
				{
					log(e, LogLevel.Fatal);
				}
			}
		}, 5000, DHTConstants.DHT_UPDATE_INTERVAL, TimeUnit.MILLISECONDS);
		started();

	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#started()
	 */
	public void started () {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#stop()
	 */
	public void stop () {
		if (!running) {
			return;
		}

		//scheduler.shutdown();
		logInfo("Stopping DHT");
		srv.stop();
		try {
			node.saveTable(table_file);
		} catch (IOException e) {
			e.printStackTrace();
		}
		running = false;
		stopped();
		tman = null;
		db = null;
		node = null;
		srv = null;
		setStatus(DHTStatus.Stopped);
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#getNode()
	 */
	public Node getNode () {
		return node;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#getTaskManager()
	 */
	public TaskManager getTaskManager () {
		return tman;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#stopped()
	 */
	public void stopped () {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#update()
	 */
	public void update () {
		if (!running) {
			return;
		}

		if (expire_timer.getElapsedSinceUpdate() > DHTConstants.CHECK_FOR_EXPIRED_ENTRIES) {
			db.expire(System.currentTimeMillis());
			expire_timer.update();
		}

		node.refreshBuckets(this);
		tman.removeFinishedTasks(this);

		if (!bootstrapping) {
			if (node.getNumEntriesInRoutingTable() < DHTConstants.BOOTSTRAP_IF_LESS_THAN_X_PEERS) {
				bootstrap();
			} else if (System.currentTimeMillis() - lastBootstrap > DHTConstants.SELF_LOOKUP_INTERVAL) {
				//update stale entries
				PingRefreshTask prt = new PingRefreshTask(srv, node, false);
				prt.setInfo("Refreshing old entries.");
				if (canStartTask()) {
					prt.start();
				}
				tman.addTask(prt);

				//regualary search for our id to update routing table
				bootstrap();
			} else {
				setStatus(DHTStatus.Running);
			}
		}

		if (hasStatsListeners()) {
			onStatsUpdate();
		}
	}

	/**
	 * Initiates a Bootstrap.
	 *
	 * This function bootstraps with router.bittorrent.com if there are less than 10 Peers in
	 * the routing table. If there are more then a lookup on our own ID is initiated.
	 * If the either Task is finished than it will try to fill the Buckets.
	 */
	public synchronized void bootstrap () {
		if (!running || bootstrapping) {
			return;
		}
		bootstrapping = true;
		logInfo("Bootstrapping...");
		lastBootstrap = System.currentTimeMillis();
		TaskListener bootstrapListener = new TaskListener() {
			/* (non-Javadoc)
			 * @see lbms.plugins.mldht.kad.TaskListener#finished(lbms.plugins.mldht.kad.Task)
			 */
			public void finished (Task t) {
				bootstrapping = false;
				if (running
						&& node.getNumEntriesInRoutingTable() > DHTConstants.USE_BT_ROUTER_IF_LESS_THAN_X_PEERS) {
					node.fillBuckets(DHT.this);
				}
			}
		};
		NodeLookup nl = findNode(node.getOurID(), true, true);
		if(nl == null) {
			bootstrapping = false;
		} else if (node.getNumEntriesInRoutingTable() < DHTConstants.USE_BT_ROUTER_IF_LESS_THAN_X_PEERS) {
			try
			{
				DHTConstants.BOOTSTRAP_NODE_ADDRESS = InetAddress.getByName(DHTConstants.BOOTSTRAP_NODE);
			} catch (Exception e)
			{
				// do nothing
			}
			nl.addDHTNode(DHTConstants.BOOTSTRAP_NODE, DHTConstants.BOOTSTRAP_PORT);
			nl.addListener(bootstrapListener);
			nl.setInfo("Bootstrap: Find Peers.");

			tman.removeFinishedTasks(this);

		} else {
			nl.setInfo("Bootstrap: search for ourself.");
			nl.addListener(bootstrapListener);
			tman.removeFinishedTasks(this);
		}

	}

	private NodeLookup findNode(Key id, boolean isBootstrap, boolean queue)
	{
		if (!running) {
			return null;
		}

		NodeLookup at = new NodeLookup(id, srv, node, isBootstrap);
		if (!queue && canStartTask()) {
			at.start();
		}
		tman.addTask(at);
		return at;
	}

	/**
	 * Do a NodeLookup.
	 * @param id The id of the key to search
	 */
	public NodeLookup findNode (Key id) {
		return findNode(id, false, false);
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.DHTBase#fillBucket(lbms.plugins.mldht.kad.KBucket)
	 */
	public NodeLookup fillBucket (Key id, KBucket bucket) {
		bucket.updateRefreshTimer();
		return findNode(id);
	}

	public PingRefreshTask refreshBucket (KBucket bucket) {
		if (!running) {
			return null;
		}

		PingRefreshTask prt = new PingRefreshTask(srv, node, bucket, false);
		if (canStartTask()) {
			prt.start();
		}
		tman.addTask(prt);

		return prt;
	}

	public boolean canStartTask () {
		// we can start a task if we have less then  7 runnning and
		// there are at least 16 RPC slots available
		if (tman.getNumTasks() >= 7) {
			return false;
		} else if (256 - srv.getNumActiveRPCCalls() <= 16) {
			return false;
		}
		return true;
	}

	public Key getOurID () {
		if (running) {
			return node.getOurID();
		}
		return null;
	}

	private boolean hasStatsListeners () {
		return !statsListeners.isEmpty();
	}

	private void onStatsUpdate () {
		stats.setNumTasks(tman.getNumTasks() + tman.getNumQueuedTasks());
		stats.setNumPeers(node.getNumEntriesInRoutingTable());
		stats.setNumSentPackets(srv.getNumSent());
		stats.setNumReceivedPackets(srv.getNumReceived());
		stats.setNumRpcCalls(srv.getNumActiveRPCCalls());

		for (int i = 0; i < statsListeners.size(); i++) {
			statsListeners.get(i).statsUpdated(stats);
		}
	}

	private void setStatus (DHTStatus status) {
		if (!this.status.equals(status)) {
			DHTStatus old = this.status;
			this.status = status;
			if (!statusListeners.isEmpty()) {
				for (int i = 0; i < statusListeners.size(); i++) {
					statusListeners.get(i).statusChanged(status, old);
				}
			}
		}
	}

	public void addStatsListener (DHTStatsListener listener) {
		statsListeners.add(listener);
	}

	public void removeStatsListener (DHTStatsListener listener) {
		statsListeners.remove(listener);
	}

	public void addStatusListener (DHTStatusListener listener) {
		statusListeners.add(listener);
	}

	public void removeStatusListener (DHTStatusListener listener) {
		statusListeners.remove(listener);
	}

	/**
	 * @return the logger
	 */
	//	public static DHTLogger getLogger () {
	//		return logger;
	//	}
	/**
	 * @param logger the logger to set
	 */
	public static void setLogger (DHTLogger logger) {
		DHT.logger = logger;
	}

	/**
	 * @return the logLevel
	 */
	public static LogLevel getLogLevel () {
		return logLevel;
	}

	/**
	 * @param logLevel the logLevel to set
	 */
	public static void setLogLevel (LogLevel logLevel) {
		DHT.logLevel = logLevel;
		logger.log("Change LogLevel to: " + logLevel);
	}

	/**
	 * @return the scheduler
	 */
	public static ScheduledExecutorService getScheduler () {
		return scheduler;
	}

	public static void log (String message, LogLevel level) {
		if (level.compareTo(logLevel) < 1) { // <=
			logger.log(message);
		}
	}

	public static void log (Exception e, LogLevel level) {
		if (level.compareTo(logLevel) < 1) { // <=
			logger.log(e);
		}
	}

	public static void logFatal (String message) {
		log(message, LogLevel.Fatal);
	}

	public static void logError (String message) {
		log(message, LogLevel.Error);
	}

	public static void logInfo (String message) {
		log(message, LogLevel.Info);
	}

	public static void logDebug (String message) {
		log(message, LogLevel.Debug);
	}

	public static void logVerbose (String message) {
		log(message, LogLevel.Verbose);
	}

	public static boolean isLogLevelEnabled (LogLevel level) {
		return level.compareTo(logLevel) < 1;
	}

	public static enum LogLevel {
		Fatal, Error, Info, Debug, Verbose
	}
}
