package lbms.plugins.mldht.kad;

import java.nio.ByteBuffer;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Damokles
 *
 */
public class Database {
	private Map<Key, List<DBItem>>	items;
	private Map<Key, Long>			tokens;
	private DatabaseStats			stats;

	Database () {
		stats = new DatabaseStats();
		items = new HashMap<Key, List<DBItem>>();
		tokens = new HashMap<Key, Long>();
	}

	/**
	 * Store an entry in the database
	 * @param key The key
	 * @param dbi The DBItem to store
	 */
	void store (Key key, DBItem dbi) {
		if (items.containsKey(key)) {
			List<DBItem> dbl = items.get(key);
			dbl = new ArrayList<DBItem>();
			items.put(key, dbl);
			dbl.add(dbi);
			stats.setItemCount(stats.getItemCount() + 1);
		}
	}

	/**
	 * Get max_entries items from the database, which have
	 * the same key, items are taken randomly from the list.
	 * If the key is not present no items will be returned, if
	 * there are fewer then max_entries items for the key, all
	 * entries will be returned
	 * @param key The key to search for
	 * @param dbl The list to store the items in
	 * @param max_entries The maximum number entries
	 */
	void sample (Key key, List<DBItem> tdbl, int max_entries) {
		if (items.containsKey(key)) {
			List<DBItem> dbl = items.get(key);

			if (dbl.size() < max_entries) {
				for (DBItem item : dbl) {
					tdbl.add(item);
				}
			} else {
				int num_added = 0;
				for (DBItem item : dbl) {
					if (num_added >= max_entries) {
						break;
					}
					tdbl.add(item);
					num_added++;
				}
			}
		}
	}

	/**
	 * Expire all items older then 30 minutes
	 * @param now The time it is now
	 * (we pass this along so we only have to calculate it once)
	 */
	void expire (long now) {
		int removed = 0;
		List<Key> toRem = new ArrayList<Key>();
		for (Key k : items.keySet()) {
			List<DBItem> dbl = items.get(k);
			// newer keys are inserted at the back
			// so we can stop when we hit the first key which is not expired
			while (dbl.size() > 0 && dbl.get(0).expired(now)) {
				dbl.remove(0);
				removed++;
			}
			if (dbl.size() == 0) {
				toRem.add(k);
			}
		}
		for (Key k : toRem) {
			items.remove(k);
		}
		stats.setKeyCount(stats.getKeyCount() - toRem.size());
		stats.setItemCount(stats.getItemCount() - removed);
	}

	/**
	 * Generate a write token, which will give peers write access to
	 * the DB.
	 * @param ip The IP of the peer
	 * @param port The port of the peer
	 * @return A Key
	 */
	Key genToken (byte[] ip, int port) {
		byte[] tdata = new byte[14];
		long now = System.currentTimeMillis();

		if (ip.length != 4) {
			System.out.println("Invalid IP");
			return null;
		}

		// generate a hash of the ip port and the current time
		// should prevent anybody from crapping things up
		ByteBuffer bb = ByteBuffer.wrap(tdata);
		bb.put(ip, 0, 4);
		bb.putShort((short) port);
		bb.putLong(now);

		try {
			MessageDigest md = MessageDigest.getInstance("SHA-1");
			md.update(tdata);
			Key token = new Key(md.digest());
			// keep track of the token, tokens will expire after a while
			tokens.put(token, now);
			return token;
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
			return Key.createRandomKey();
		}
	}

	/**
	 * Check if a received token is OK.
	 * @param token The token received
	 * @param ip The ip of the sender
	 * @param port The port of the sender
	 * @return true if the token was given to this peer, false other wise
	 */
	boolean checkToken (Key token, byte[] ip, int port) {

		if (ip.length != 4) {
			return false;
		}

		// the token must be in the map
		if (!tokens.containsKey(token)) {
			DHT.logDebug("Received Unknown token from " + (ip[0] & 0xFF) + "."
					+ (ip[1] & 0xFF) + "." + (ip[2] & 0xFF) + "."
					+ (ip[3] & 0xFF));
			return false;
		}

		// in the map so now get the timestamp and regenerate the token
		// using the IP and port of the sender
		long ts = tokens.get(token);
		byte[] tdata = new byte[14];
		ByteBuffer bb = ByteBuffer.wrap(tdata);
		bb.put(ip, 0, 4);
		bb.putShort((short) port);
		bb.putLong(ts);
		try {
			MessageDigest md = MessageDigest.getInstance("SHA-1");
			md.update(tdata);

			Key ct = new Key(md.digest());
			// compare the generated token to the one received
			if (token != ct) // not good, this peer didn't went through the proper channels
			{
				DHT.logDebug("Received Invalid token from " + (ip[0] & 0xFF)
						+ "." + (ip[1] & 0xFF) + "." + (ip[2] & 0xFF) + "."
						+ (ip[3] & 0xFF));
				return false;
			}
			// expire the token
			tokens.remove(token);
			return true;
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
			return false;
		}
	}

	/// Test wether or not the DB contains a key
	boolean contains (Key key) {
		return items.containsKey(key);
	}

	/// Insert an empty item (only if it isn't already in the DB)
	void insert (Key key) {
		List<DBItem> dbl = items.get(key);
		if (dbl == null) {
			dbl = new ArrayList<DBItem>();
			items.put(key, dbl);
			stats.setKeyCount(stats.getKeyCount() + 1);
		}
	}

	/**
	 * @return the stats
	 */
	public DatabaseStats getStats () {
		return stats;
	}
}
