package lbms.plugins.mldht.kad;

import java.util.Collection;
import java.util.Collections;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * @author Damokles
 *
 */
public class KClosestNodesSearch {
	private Key								key;
	private SortedMap<Key, KBucketEntry>	emap;
	private int								max_entries;

	/**
	 * Constructor sets the key to compare with
	 * @param key The key to compare with
	 * @param max_entries The maximum number of entries can be in the map
	 * @return
	 */
	public KClosestNodesSearch (Key key, int max_entries) {
		this.key = key;
		this.max_entries = max_entries;
		this.emap = new TreeMap<Key, KBucketEntry>();
	}

	/**
	 * @return the Target key of the search
	 */
	public Key getSearchTarget () {
		return key;
	}

	/**
	 * @return the number of entries
	 */
	public int getNumEntries () {
		return emap.size();
	}

	/**
	 * Try to insert an entry.
	 * @param e The entry
	 * @return true if insert was successful
	 */
	public boolean tryInsert (KBucketEntry e) {
		// calculate distance between key and e
		Key d = key.distance(e.getID());

		if (emap.size() < max_entries) {
			// room in the map so just insert
			emap.put(d, e);
			return true;
		} else {
			// now find the max distance
			// seeing that the last element of the map has also
			// the biggest distance to key (std::map is sorted on the distance)
			// we just take the last
			Key max = emap.lastKey();
			if (d.compareTo(max) == -1) {
				// insert if d is smaller then max
				emap.put(d, e);
				// erase the old max value
				emap.remove(max);
				return true;
			}
			return false;
		}
	}

	public boolean isFull () {
		return emap.size() >= max_entries;
	}

	/**
	 * Packs the results in a byte array.
	 *
	 * @return the encoded results.
	 */
	public byte[] pack () {
		byte[] buffer = new byte[emap.size() * 26];
		int max_items = buffer.length / 26;
		int j = 0;

		for (KBucketEntry e : emap.values()) {
			if (max_items >= j) {
				break;
			}
			PackUtil.PackBucketEntry(e, buffer, j * 26);
			j++;
		}
		return buffer;
	}

	/**
	 * @return a unmodifiable Collection of the entries
	 */
	public Collection<KBucketEntry> getEntries () {
		return Collections.unmodifiableCollection(emap.values());
	}
}
