package lbms.plugins.mldht.kad;

import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;
import java.util.SortedSet;

import lbms.plugins.mldht.kad.DHT.LogLevel;

/**
 * @author The_8472, Damokles
 *
 */
public class PopulationEstimator {

	static final int						KEYSPACE_BITS					= Key.SHA1_HASH_LENGTH * 8;
	static final double						KEYSPACE_SIZE					= Math.pow(2, KEYSPACE_BITS);
	
	// try to close in to a rough estimate fast
	static final double						DISTANCE_INITIAL_WEIGHT			= 0.1;									// 10%
	static final int						INITIAL_WEIGHT_COUNT			= 45;
	// apply less weight to individual estimates later on
	static final double						DISTANCE_INCREASE_WEIGHT		= 0.007;								// 0.8%
	static final double						DISTANCE_DECREASE_WEIGHT		= 0.03;								// 3%
	
	private static double					averageNodeDistance				= KEYSPACE_SIZE;
	private static int						updateCount						= 0;
	private static List<PopulationListener>	listeners						= new ArrayList<PopulationListener>(1);
	
	private static int						MAX_RECENT_LOOKUP_CACHE_SIZE	= 40;
	private static Deque<Key>				recentlySeenPrefixes			= new LinkedList<Key>();

	static public long getEstimate () {
		return (long) (KEYSPACE_SIZE / averageNodeDistance);
	}
	
	static double getRawDistanceEstimate() {
		return averageNodeDistance;
	}

	static void setInitialDistanceEstimate(double initialValue) {
		averageNodeDistance = initialValue;
	}

	static public void update (SortedSet<Key> neighbors) {
		if(neighbors.size() < 2)
			return;
		DHT.log("Estimator: new node group of "+neighbors.size(), LogLevel.Debug);
		Key prefix = Key.getCommonPrefix(neighbors);
		
		synchronized (recentlySeenPrefixes)
		{
			for(Key oldPrefix : recentlySeenPrefixes)
			{
				if(prefix.isPrefixOf(oldPrefix) || oldPrefix.isPrefixOf(prefix))
				{
					/*
					 * displace old entry, narrower entries will also replace
					 * wider ones, to clean out accidents like prefixes covering
					 * huge fractions of the keyspace
					 */
					recentlySeenPrefixes.remove(oldPrefix);
					recentlySeenPrefixes.addLast(prefix);
					return;
				}
			}

			// no match found => add
			recentlySeenPrefixes.addLast(prefix);
			if(recentlySeenPrefixes.size() > MAX_RECENT_LOOKUP_CACHE_SIZE)
				recentlySeenPrefixes.removeFirst();
		}
		
		Key previous = null;
		for (Key entry : neighbors) {
			if (previous == null) {
				previous = entry;
				continue;
			}

			byte[] rawDistance = previous.distance(entry).getHash();

			double distance = 0;

			int nonZeroBytes = 0;
			for (int i = 0; i < Key.SHA1_HASH_LENGTH; i++) {
				if (rawDistance[i] == 0) {
					continue;
				}
				if (nonZeroBytes == 3) {
					break;
				}
				nonZeroBytes++;
				distance += (rawDistance[i] & 0xFF)
						* Math.pow(2, KEYSPACE_BITS - (i + 1) * 8);
			}

			/*
			 * weighted average of the logarithms (since single results can be
			 * off by several orders of magnitude -> logarithm dampens that
			 * exponentially) prefer distance decreases as we'll naturally hit a
			 * lower bound there (on average) whereas we can always get larger
			 * distances due to errors in the lookup
			 */
			
			double weight;
			if (distance < averageNodeDistance)
				weight = DISTANCE_DECREASE_WEIGHT;
			else
				weight = DISTANCE_INCREASE_WEIGHT;
			if (updateCount < INITIAL_WEIGHT_COUNT)
				weight = DISTANCE_INITIAL_WEIGHT;
			synchronized (PopulationEstimator.class)
			{
				averageNodeDistance = Math.exp(Math.log(distance) * weight + Math.log(averageNodeDistance) * (1. - weight));
			}
			
			updateCount++;
			DHT.log("Estimator: distance update #"+updateCount+": " + distance + " avg:" + averageNodeDistance, LogLevel.Debug);

			previous = entry;
		}
		
		DHT.log("Estimator: new estimate:"+getEstimate(), LogLevel.Info);
		
		fireUpdateEvent();
	}

	public static void addListener (PopulationListener l) {
		listeners.add(l);
	}

	public static void removeListener (PopulationListener l) {
		listeners.remove(l);
	}

	private static void fireUpdateEvent () {
		long estimated = getEstimate();
		for (int i = 0; i < listeners.size(); i++) {
			listeners.get(i).populationUpdated(estimated);
		}
	}
}
