package lbms.plugins.mldht.kad;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import lbms.plugins.mldht.kad.DHT.LogLevel;
import lbms.plugins.mldht.kad.messages.MessageBase;
import lbms.plugins.mldht.kad.messages.MessageDecoder;
import lbms.plugins.mldht.kad.messages.PingRequest;
import lbms.plugins.mldht.kad.messages.MessageBase.Type;

import org.gudy.azureus2.core3.util.BDecoder;

/**
 * @author Damokles
 *
 */
public class RPCServer implements Runnable, RPCServerBase {

	private DatagramSocket				sock;
	private DHT							dh_table;
	private Map<Integer, RPCCallBase>	calls;
	private List<RPCCallBase>			call_queue;
	private byte						next_mtid;
	private volatile boolean			running;
	private Thread						thread;

	private int							numReceived;
	private int							numSent;
	private RPCStats					stats;

	public RPCServer (DHT dh_table, int port) throws SocketException { //, Object parent
		sock = new DatagramSocket(port);
		calls = new HashMap<Integer, RPCCallBase>();
		call_queue = new ArrayList<RPCCallBase>(128);
		stats = new RPCStats();
		this.dh_table = dh_table;
	}

	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#run()
	 */
	public void run () {
		while (running && !sock.isClosed()) {
			byte[] buffer = new byte[DHTConstants.RECEIVE_BUFFER_SIZE];
			DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
			try {
				sock.receive(packet);
				handlePacket(packet);
			} catch (Exception e) {
				if(running)
					DHT.log(e, LogLevel.Error);
			}
		}
		DHT.logInfo("Stopped RPC Server");
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#start()
	 */
	public void start () {
		DHT.logInfo("Starting RPC Server");
		running = true;
		thread = new Thread(this, "mlDHT RPC Thread");
		thread.setPriority(Thread.MIN_PRIORITY);
		thread.setDaemon(true);
		thread.start();
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#stop()
	 */
	public void stop () {
		running = false;
		sock.close();
		thread = null;
		DHT.logInfo("Stopping RPC Server");
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#doCall(lbms.plugins.mldht.kad.messages.MessageBase)
	 */
	public RPCCall doCall (MessageBase msg) {
		byte start = next_mtid;
		while (calls.containsKey(next_mtid & 0xFF)) {
			next_mtid++;
			if (next_mtid == start) // if this happens we cannot do any calls
			{
				// so queue the call
				RPCCall c = new RPCCall(this, msg, true);
				call_queue.add(c);
				System.out.println("Queueing RPC call, no slots available at the moment");
				return c;
			}
		}

		msg.setMTID(next_mtid++);
		sendMessage(msg);
		RPCCall c = new RPCCall(this, msg, false);
		calls.put(msg.getMTID() & 0xFF, c);
		return c;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#timedOut(byte)
	 */
	public void timedOut (byte mtid) {
		// delete the call
		RPCCallBase c = calls.get(mtid & 0xFF);
		if (c != null) {
			stats.addTimeoutMessageToCount(c.getRequest());
			dh_table.timeout(c.getRequest());
			calls.remove(mtid & 0xFF);
		}
		doQueuedCalls();
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#ping(lbms.plugins.mldht.kad.Key, java.net.InetSocketAddress)
	 */
	public void ping (Key our_id, InetSocketAddress addr) {
		PingRequest pr = new PingRequest(our_id);
		pr.setOrigin(addr);
		doCall(pr);
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#findCall(byte)
	 */
	public RPCCallBase findCall (byte mtid) {
		return calls.get(mtid & 0xFF);
	}

	/// Get the number of active calls
	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#getNumActiveRPCCalls()
	 */
	public int getNumActiveRPCCalls () {
		return calls.size();
	}

	/**
	 * @return the numReceived
	 */
	public int getNumReceived () {
		return numReceived;
	}

	/**
	 * @return the numSent
	 */
	public int getNumSent () {
		return numSent;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#getStats()
	 */
	public RPCStats getStats () {
		return stats;
	}

	private void handlePacket (DatagramPacket p) {
		numReceived++;
		stats.addReceivedBytes(p.getLength());
		byte[] ip = p.getAddress().getAddress();
		DHT.logDebug("RPC received Packet from: " + (ip[0] & 0xFF) + "."
				+ (ip[1] & 0xFF) + "." + (ip[2] & 0xFF) + "." + (ip[3] & 0xFF));
		if (DHT.isLogLevelEnabled(LogLevel.Verbose)) {
			try {
				DHT.logVerbose(new String(p.getData(), 0, p.getLength(),
						"UTF-8"));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		try {
			byte[] buf = new byte[p.getLength()];
			System.arraycopy(p.getData(), p.getOffset(), buf, 0, buf.length);
			Map<String, Object> bedata = BDecoder.decode(buf);
			MessageBase msg = MessageDecoder.parseMessage(bedata, this);
			if (msg != null) {
				DHT.logDebug("RPC Packet was: " + msg.getMethod() + " Type:"
						+ msg.getType() + " MessageID: "
						+ (msg.getMTID() & 0xFF));
				stats.addReceivedMessageToCount(msg);
				msg.setOrigin(new InetSocketAddress(p.getAddress(), p.getPort()));
				msg.apply(dh_table);
				// erase an existing call
				if (msg.getType() == Type.RSP_MSG
						&& calls.containsKey(msg.getMTID() & 0xFF)) {
					// delete the call, but first notify it off the response
					RPCCallBase c = calls.get(msg.getMTID() & 0xFF);
					c.response(msg);
					calls.remove(msg.getMTID() & 0xFF);
					doQueuedCalls();
				}
			}
		} catch (IOException e) {
			DHT.log(e, LogLevel.Debug);
		}
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#sendMessage(lbms.plugins.mldht.kad.messages.MessageBase)
	 */
	public void sendMessage (MessageBase msg) {
		try {
			stats.addSentMessageToCount(msg);
			send(msg.getDestination(), msg.encode());
			DHT.logDebug("RPC Send Message: " + msg.getMethod() + " Type:"
					+ msg.getType() + " MessageID: " + (msg.getMTID() & 0xFF));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void send (InetSocketAddress addr, byte[] msg) throws IOException {
		if (!sock.isClosed()) {
			DatagramPacket p = new DatagramPacket(msg, msg.length);
			p.setAddress(addr.getAddress());
			p.setPort(addr.getPort());
			sock.send(p);
			stats.addSentBytes(msg.length);
			numSent++;
		}
	}

	private void doQueuedCalls () {
		while (call_queue.size() > 0 && calls.size() < 256) {
			RPCCallBase c = call_queue.remove(0);

			while (calls.containsKey(next_mtid & 0xFF)) {
				next_mtid++;
			}

			MessageBase msg = c.getRequest();
			msg.setMTID(next_mtid++);
			sendMessage(msg);
			calls.put(msg.getMTID() & 0xFF, c);
			c.start();
		}
	}

}
